#include <vector>
#include <set>
#include <stdexcept>

using namespace std;

#include <json/json.h>

#include <WorldBase.hh>
#include <Entity.hh>
#include <tools.hh>

using namespace HATP_TOOLS;

#include <ILAP.hh>

std::string assemblyEntityNameMappingToHATP(const std::string& name);

#define ASSEMBLY_PLANNER_NAME "AssemblyPlanner"
//#define ASSEMBLY_PLANNER_NAME "AssemblyFilter"

bool ILAP::connect(const std::string& server, unsigned int port){
	if(!client.connect("ILAP", server, port)){
		return false;
	}
	
	//Before continuing, clean the stock of messages
	while(client.isMessageWaiting()){
		std::pair<std::string, std::string> garbage;
		garbage=client.getMessage();
	}
	
	return true;
}

//Turn the name into a valid HATP name
//Entity name must be composed as follow:
//('A'..'Z') ('a'..'z' | 'A'..'Z' | '0'..'9'| '_')*
//
//This function automatically replace ' ' and '.', by .
void ILAP::validateName(std::string& name) const{
	const std::ctype<char>& facet(std::use_facet< std::ctype<char> >(std::locale()));
	
	//Change the name using the aliases defined by the user
//	std::cout<<"Name before1: "<<name<<std::endl;
	name=assemblyEntityNameMappingToHATP(name);
//	std::cout<<"Name after1: "<<name<<std::endl;
	
	//Check that its frist letter is a letter and change it to an upper case
	if(!facet.is(std::ctype<char>::alpha, name[0])){
		throw std::invalid_argument("Invalid name \""+name+"\", first symbol must be a letter");
	}
	name[0]=facet.toupper(name[0]);
	
	//Change the '.' and ' ' into underscores
	for(size_t pos=1; pos<name.size(); ++pos){
		if(name[pos]=='.' || name[pos]==' '){
			name[pos]='_';
			continue;
		}
		if(!facet.is(std::ctype<char>::alnum, name[pos]) &&
			name[pos]!='_'){
			throw std::invalid_argument("Invalid name \""+name+"\", invalid symbol \'"+name[pos]+"\'");
		}
	}
	
//	std::cout<<"Name before2: "<<name<<std::endl;
//	name=assemblyEntityNameMappingToHATP(name);
//	std::cout<<"Name after2: "<<name<<std::endl;
}

std::string ILAP::requestPlan(std::string const& structure){
	std::stringstream ss_req;
	ss_req<<"{\"assembly-request\": { \"request-type\": \"plan\", \"request-id\": "<<lastID<<", \"parameters\": [{\"structure\":\""<<structure<<"\"}]}} ";
//	ss_req<<"{\"assembly-request\": { \"request-type\": \"plan\", \"request-id\": "<<lastID<<", \"parameters\": [{\"structure\":\"cross.DAE\"}]}} ";

	++lastID;
	std::string request(ss_req.str());
	client.sendMessage(ASSEMBLY_PLANNER_NAME, request);

	std::cout<<"Request sent to Assembly planner, waiting for the plan ..."<<std::endl;
	std::pair<std::string, std::string> result=client.getBlockingMessage();
	if(result.first.empty()){
		std::cerr<<"ERROR: Problem when requesting the assembly plan :"<<result.second<<std::endl;
		return "";
	}
	if(result.first!=ASSEMBLY_PLANNER_NAME){
		std::cerr<<"ERROR: Receiving answer from \""<<result.first<<"\" instead of \""<<ASSEMBLY_PLANNER_NAME<<"\""<<std::endl;
		return "";
	}
	std::string toParse=result.second;

	return toParse;
}

std::string ILAP::requestREPlan(){
	std::stringstream ss_req;
	if(!error_on_part.empty()){
		unsigned int part_id;
		for(std::map<unsigned int, std::string>::const_iterator it=pieces.begin(); it!=pieces.end(); ++it){
			if(it->second==error_on_part){
				part_id=it->first;
				break;
			}
		}
		ss_req<<"{\"assembly-request\": { \"request-type\": \"replan\", \"request-id\": "<<(lastID-1)<<", \"parameters\": [{\"error-on-part\":"<<part_id<<"}]}}";

	}else{
		ss_req<<"{\"assembly-request\": { \"request-type\": \"replan\", \"request-id\": "<<(lastID-1)<<", \"parameters\": []}}";
	}
	
	//TODO TELL CARMELLO ABOUT THE ID THAT TRIGGERS AN ASSERTION - TWO STRING FOR A COMMON ERROR
	std::string request(ss_req.str());
	client.sendMessage(ASSEMBLY_PLANNER_NAME, request);

	std::cout<<"Request sent to Assembly planner, waiting for the REplan ..."<<std::endl;
	std::pair<std::string, std::string> result=client.getBlockingMessage();
	if(result.first.empty()){
		std::cerr<<"ERROR: Problem when requesting the assembly plan :"<<result.second<<std::endl;
		return "";
	}
	if(result.first!=ASSEMBLY_PLANNER_NAME){
		std::cerr<<"ERROR: Receiving answer from \""<<result.first<<"\" instead of \""<<ASSEMBLY_PLANNER_NAME<<"\""<<std::endl;
		return "";
	}
	std::string toParse=result.second;

	return toParse;
}

bool ILAP::updateWorldBaseWithAssemblyPlan(WorldBase& wb, Entity* structure){ //TODO, std::string const& requested_structure){
//	std::string requested_structure("equivalent.struct.dae");
	std::string requested_structure(structure->get<std::string>("file"));
	std::string toParse(requestPlan(requested_structure));
	
	if(toParse.empty()){
		if(!client.isConnected()){
			std::cerr<<"ERROR: ILAP is not connected"<<std::endl;
			return false;
		}
		std::cerr<<"ERROR: Could not get a proper anser to assembly plan request"<<std::endl;
		return false;
	}
	
	Json::Value planJSON;
	Json::Reader reader;
	if(!reader.parse(toParse, planJSON)){
		std::cerr<<"ERROR: Invalid message"<<std::endl;
		std::cerr<<toParse<<std::endl;
		return false;
	}

	//Structure name
//	structure = planJSON.get("name", "Structure").asString();
//	validateName(structure);

	//Analyses the pieces
	Json::Value piecesJSON = planJSON["parts"];
	if(piecesJSON.isNull() || !piecesJSON.isArray()){
		throw std::invalid_argument("Invalid Assembly Plan, does not contain the list of pieces");
	}
	
//	std::map<unsigned int, std::string> pieces;
	pieces.clear();
	for(Json::Value::ArrayIndex i=0; i<piecesJSON.size(); ++i){
		if(piecesJSON[(int)i]["id"].isNull() || !piecesJSON[i]["id"].isIntegral() ||
			piecesJSON[i]["name"].isNull() || !piecesJSON[i]["name"].isString()){
			std::cerr<<"ERROR: The piece is invalid:"<<piecesJSON[i].toStyledString()<<std::endl;
			return false;
		}
		unsigned int id=piecesJSON[i]["id"].asUInt();
		std::string piece=piecesJSON[i]["name"].asString();
		validateName(piece);
		if(pieces.insert(std::make_pair(id, piece)).second==false){
			std::cerr<<"ERROR: The piece "<<piece<<" has already been declared"<<std::endl;
			return false;
		}
	}
	
	//Analyse the operations
	Json::Value operationsJSON = planJSON["operations"];
	if(operationsJSON.isNull() || !operationsJSON.isArray()){
		throw std::invalid_argument("Invalid Assembly Plan, does not contain the list of operations");
	}
	
	std::map<std::string, std::string> supports; //Pair<Piece, Support>
	std::map<std::string, std::set<std::string> > dependencies; //Pair<Piece, set<dependencies>>
	std::map<std::string, std::string> predecessors; //Pair<Piece, Predecessor>
	std::string predecessor;
	for(Json::Value::ArrayIndex i=0; i<operationsJSON.size(); ++i){
		if(operationsJSON[i]["depends"].isNull() || !operationsJSON[i]["depends"].isArray() ||
			operationsJSON[i]["pieces"].isNull() || !operationsJSON[i]["pieces"].isArray() ||
			!operationsJSON[i]["pieces"].size()==1 || !operationsJSON[i]["pieces"][0].isIntegral() ||
			operationsJSON[i]["support"].isNull() || !operationsJSON[i]["support"].isString() ||
			operationsJSON[i]["type"].isNull() || !operationsJSON[i]["type"].isString()){
			std::cerr<<"ERROR: The operation is invalid:"<<operationsJSON[i].toStyledString()<<std::endl;
			return false;
		}
		
		//Detect the piece
		unsigned int pieceId=operationsJSON[i]["pieces"][0].asUInt();
		if(pieces.find(pieceId)==pieces.end()){
			std::cerr<<"ERROR: The piece "<<pieceId<<" does not exist"<<std::endl;
			return false;
		}
		std::string piece=pieces.find(pieceId)->second;

		//Create a pair with the piece and its support
//TODO		std::string support=operationsJSON[i]["support"].asString();
		std::string support="Support_"+piece;
		validateName(support);
		if(supports.insert(std::make_pair(piece, support)).second==false){
			std::cerr<<"ERROR: The piece "<<piece<<" already has a support"<<std::endl;
			return false;
		}
		
		//Add the dependencies
		Json::Value dependenciesJSON = operationsJSON[i]["depends"];
		for(Json::Value::ArrayIndex j=0; j<dependenciesJSON.size(); ++j){
			if(!dependenciesJSON[j].isIntegral() ||
				pieces.find(dependenciesJSON[j].asUInt())==pieces.end()){
				std::cerr<<"ERROR: Invalid dependencies for "<<pieceId<<std::endl;
				return false;
			}
			std::string dependency=pieces.find(dependenciesJSON[j].asUInt())->second;
			if(dependencies[piece].insert(dependency).second==false){
				std::cerr<<"ERROR: The piece "<<piece<<" already has a dependency to "<<dependency<<std::endl;
				return false;
			}
		}
		
		//Handle the order of the operations
		if(!predecessor.empty()){
			if(predecessors.insert(std::make_pair(piece, predecessor)).second==false){
				std::cerr<<"ERROR: The piece "<<piece<<" already has a predecessor, impossible to add "<<predecessor<<std::endl;
				return false;
			}
		}
		predecessor=piece;
	}
	
	//Now update the WorldBase
//	wb.add(entityFactory(structure, "Structure"));
	
	std::map<unsigned int, std::string>::const_iterator itPiece;
	for(itPiece=pieces.begin(); itPiece!=pieces.end(); ++itPiece){
		wb.add(entityFactory(itPiece->second, "Object"));
//		wb.find(structure)->get< set<Entity*> >("parts") <<= wb.find(itPiece->second);
		wb.find(structure->getName())->get< set<Entity*> >("parts") <<= wb.find(itPiece->second);
	}
	
	std::map<std::string, std::string>::const_iterator itSupport;
	for(itSupport=supports.begin(); itSupport!=supports.end(); ++itSupport){
		wb.add(entityFactory(itSupport->second, "Support"));
		wb.find(itSupport->first)->get<Entity*>("support") = wb.find(itSupport->second);
	}
	
	std::map<std::string, std::set<std::string> >::const_iterator itDependency;
	for(itDependency=dependencies.begin(); itDependency!=dependencies.end(); ++itDependency){
		std::set<std::string> const& list=itDependency->second;
		std::set<std::string>::const_iterator itList;
		for(itList=list.begin(); itList!=list.end(); ++itList){
			wb.find(itDependency->first)->get< set<Entity*> >("dependsOn") <<= wb.find(*itList);
		}
	}
	
	std::map<std::string, std::string>::const_iterator itPredecessor;
	for(itPredecessor=predecessors.begin(); itPredecessor!=predecessors.end(); ++itPredecessor){
		wb.find(itPredecessor->first)->get<Entity*>("predecessor") = wb.find(itPredecessor->second);
	}
	
	return true;
}

//bool ILAP::REPLAN_updateWorldBaseWithAssemblyPlan(WorldBase& wb){
//	std::string toParse(requestREPlan());
//	
//	if(toParse.empty()){
//		if(!client.isConnected()){
//			std::cerr<<"ERROR: ILAP is not connected"<<std::endl;
//			return false;
//		}
//		std::cerr<<"ERROR: Could not get a proper anser to assembly plan request"<<std::endl;
//		return false;
//	}
//	
//	Json::Value planJSON;
//	Json::Reader reader;
//	if(!reader.parse(toParse, planJSON)){
//		std::cerr<<"ERROR: Invalid message"<<std::endl;
//		std::cerr<<toParse<<std::endl;
//		return false;
//	}

//	//Analyse the operations
//	Json::Value operationsJSON = planJSON["operations"];
//	if(operationsJSON.isNull() || !operationsJSON.isArray()){
//		throw std::invalid_argument("Invalid Assembly Plan, does not contain the list of operations");
//	}
//	
//	std::map<std::string, std::string> supports; //Pair<Piece, Support>
//	std::map<std::string, std::set<std::string> > dependencies; //Pair<Piece, set<dependencies>>
//	std::map<std::string, std::string> predecessors; //Pair<Piece, Predecessor>
//	std::string predecessor;
//	for(Json::Value::ArrayIndex i=0; i<operationsJSON.size(); ++i){
//		if(operationsJSON[i]["depends"].isNull() || !operationsJSON[i]["depends"].isArray() ||
//			operationsJSON[i]["pieces"].isNull() || !operationsJSON[i]["pieces"].isArray() ||
//			!operationsJSON[i]["pieces"].size()==1 || !operationsJSON[i]["pieces"][0].isIntegral() ||
//			operationsJSON[i]["support"].isNull() || !operationsJSON[i]["support"].isString() ||
//			operationsJSON[i]["type"].isNull() || !operationsJSON[i]["type"].isString()){
//			std::cerr<<"ERROR: The operation is invalid:"<<operationsJSON[i].toStyledString()<<std::endl;
//			return false;
//		}
//		
//		//Detect the piece
//		unsigned int pieceId=operationsJSON[i]["pieces"][0].asUInt();
//		if(pieces.find(pieceId)==pieces.end()){
//			std::cerr<<"ERROR: The piece "<<pieceId<<" does not exist"<<std::endl;
//			return false;
//		}
//		std::string piece=pieces.find(pieceId)->second;

//		//Create a pair with the piece and its support
//		std::string support=operationsJSON[i]["support"].asString();
//		validateName(support);
//		if(supports.insert(std::make_pair(piece, support)).second==false){
//			std::cerr<<"ERROR: The piece "<<piece<<" already has a support"<<std::endl;
//			return false;
//		}
//		
//		//Add the dependencies
//		Json::Value dependenciesJSON = operationsJSON[i]["depends"];
//		for(Json::Value::ArrayIndex j=0; j<dependenciesJSON.size(); ++j){
//			if(!dependenciesJSON[j].isIntegral() ||
//				pieces.find(dependenciesJSON[j].asUInt())==pieces.end()){
//				std::cerr<<"ERROR: Invalid dependencies for "<<pieceId<<std::endl;
//				return false;
//			}
//			std::string dependency=pieces.find(dependenciesJSON[j].asUInt())->second;
//			if(dependencies[piece].insert(dependency).second==false){
//				std::cerr<<"ERROR: The piece "<<piece<<" already has a dependency to "<<dependency<<std::endl;
//				return false;
//			}
//		}
//		
//		//Handle the order of the operations
//		if(!predecessor.empty()){
//			if(predecessors.insert(std::make_pair(piece, predecessor)).second==false){
//				std::cerr<<"ERROR: The piece "<<piece<<" already has a predecessor, impossible to add "<<predecessor<<std::endl;
//				return false;
//			}
//		}
//		predecessor=piece;
//	}
//	
//	//Now update the WorldBase (suppose the workd base has been cleansed)
//	wb.add(entityFactory(structure, "Structure"));
//	
//	std::map<unsigned int, std::string>::const_iterator itPiece;
//	for(itPiece=pieces.begin(); itPiece!=pieces.end(); ++itPiece){
//		wb.add(entityFactory(itPiece->second, "Object"));
//		wb.find(structure)->get< set<Entity*> >("parts") <<= wb.find(itPiece->second);
//	}
//	
//	std::map<std::string, std::string>::const_iterator itSupport;
//	for(itSupport=supports.begin(); itSupport!=supports.end(); ++itSupport){
//		wb.add(entityFactory(itSupport->second, "Support"));
//		wb.find(itSupport->first)->get<Entity*>("support") = wb.find(itSupport->second);
//	}
//	
//	std::map<std::string, std::set<std::string> >::const_iterator itDependency;
//	for(itDependency=dependencies.begin(); itDependency!=dependencies.end(); ++itDependency){
//		std::set<std::string> const& list=itDependency->second;
//		std::set<std::string>::const_iterator itList;
//		for(itList=list.begin(); itList!=list.end(); ++itList){
//			wb.find(itDependency->first)->get< set<Entity*> >("dependsOn") <<= wb.find(*itList);
//		}
//	}
//	
//	std::map<std::string, std::string>::const_iterator itPredecessor;
//	for(itPredecessor=predecessors.begin(); itPredecessor!=predecessors.end(); ++itPredecessor){
//		wb.find(itPredecessor->first)->get<Entity*>("predecessor") = wb.find(itPredecessor->second);
//	}
//	
//	return true;
//}
