#include <tools.hh>
using namespace HATP_TOOLS;

#include <WorldBase.hh>
#include <LightWorldBase.hh>
#include <Entity.hh>

#include <ILGTP.hh>

#include <iostream>
#include <string>

ILGTP::ILGTP():
	useGetSymbolicFacts(true)
{}

ILGTP::~ILGTP(){
	disconnectGTP();
}

/** Connect to geometric task planner (GTP) through TCL
 */
bool ILGTP::connectGTP(const std::string& server, unsigned int port, unsigned short VERBOSE_LEVEL) {
	this->VERBOSE_LEVEL=VERBOSE_LEVEL;

#ifndef MANUAL_DEBUG_MODE
	if(client.isConnected())
		return true;

	VERBOSE_2 << "Connecting to MsgServer:" << server << ":" << port << std::endl;
	client.connect("ILGTP", server, port);

	if(!client.isConnected()) {
		VERBOSE_2 << "ERROR: Could not connect to MsgServer." << std::endl;
		return false;
	}

	VERBOSE_2 << "Connected to MsgServer." << std::endl;

	VERBOSE_2 << "Connecting to Move3D module..." << std::endl;

	Json::Value request;
	std::pair<std::string, std::string> result;

	request.clear();
	request["PING"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getMessageTimeout(5000);
	checkStatus(result, VERBOSE_LEVEL);
#endif //MANUAL_DEBUG_MODE
	VERBOSE_2 << "Connected" << std::endl;

	return true;
}

/** Disconnect from geometric task planner
*/
bool ILGTP::disconnectGTP() {
#ifndef MANUAL_DEBUG_MODE
	if(!client.isConnected())
		return true;

	VERBOSE_2 << "Disconnecting from MsgServer and Move3D module..." << std::endl;
	client.disconnect();
#endif //MANUAL_DEBUG_MODE
	VERBOSE_2 << "Correctly disconnected." << std::endl;

	return true;
}

bool ILGTP::initGTP(WorldBase& wb){
#ifndef MANUAL_DEBUG_MODE
	checkClient();

	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();

	//Set the environment to its default position
	request.clear();
	request["InitGTP"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);

	//Set the root of the GTP tree
	request.clear();
	request["SetGTPInitState"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);

	//Disable all unnecessary facts before requesting them, and (re)enable only the needed
	request.clear();
	request["DisableAllGTPFact"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);

	//Re-enable the facts to use
	std::vector<std::string> factsToUse(gtpFactsToUse());
	if(!factsToUse.empty()){
		for(std::vector<std::string>::const_iterator it=factsToUse.begin(); it!=factsToUse.end(); ++it){
			request.clear();
			request["DisableGTPFact"]["factType"]=*it;
			request["DisableGTPFact"]["value"]=false; //Calling Disable(false) = enable
			client.sendMessage(move3D, WRITE(request));
			result = client.getBlockingMessage();
			checkStatus(result, VERBOSE_LEVEL);
		}
	}else{
		useGetSymbolicFacts=false;
	}

	//Get the initial GTP state
	return this->getSymbolicFacts(wb, true, std::make_pair(-1, -1));
#endif //MANUAL_DEBUG_MODE
	return true;
}

/** Projects an action on GTP planner
 * The parameters for the task (the agent, the object, ...) must have been set before
 */
bool ILGTP::projectActionGTP(const std::string& taskName, TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();

	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();

	std::string requestType;

	if(!instance.isGTPAlternative()){
		requestType="PlanGTPTask";
		request[requestType]["type"]=taskName;
		request[requestType]["computeMotionPlan"]=true;

		TaskInstance* predecessor=session.getPredecessorInstance(instance);
		if(predecessor==NULL){
			request[requestType]["previousTaskId"]=-1;
			request[requestType]["previousTaskAlternativeId"]=-1;
		}else{
			request[requestType]["previousTaskId"]=predecessor->getGTPProjection().first;
			request[requestType]["previousTaskAlternativeId"]=predecessor->getGTPProjection().second;

		}

		client.sendMessage(move3D, WRITE(request));
		result = client.getBlockingMessage();
	}else{
		requestType="FindGTPTaskAlternative";
		if(instance.getGTPOriginal()==NULL){
			VERBOSE_2<<"ERROR: This task is an alternative but it does not have an original"<<std::endl;
		}

		request[requestType]["taskId"]=instance.getGTPOriginal()->getGTPProjection().first;
		request[requestType]["computeMotionPlan"]=true;

		client.sendMessage(move3D, WRITE(request));
		result = client.getBlockingMessage();
	}

	Json::Value answer;
	Json::Reader reader;
	if(!reader.parse(result.second, answer)){
		VERBOSE_3<<"ERROR: Invalid format for answer"<<std::endl;
		return false;
	}

	if(answer[requestType].isNull() || 
		!answer[requestType]["returnReport"].isString()){
		VERBOSE_3<<"ERROR: Invalid answer to planning request"<<std::endl;
		return false;
	}

	bool status;
	int newTaskId, newTaskAlternativeId;
	double newCost;
	double planningTime=0.;
	double numberOfAttempt=0.;

	status=(answer[requestType]["returnReport"].asString()=="OK");
	if(status){
		newCost=answer[requestType]["cost"].asDouble();
		newTaskId=answer[requestType]["taskId"].asInt();
		newTaskAlternativeId=answer[requestType]["taskAlternativeId"].asInt();
		
		if(!answer[requestType]["planningInformation"].isNull()){
			planningTime=answer[requestType]["planningInformation"]["planningTime"].asDouble();
			numberOfAttempt=answer[requestType]["planningInformation"]["numberOfAttempt"].asDouble();
		}
	}else{
		newCost=-1;
		newTaskId=-1;
		newTaskAlternativeId=-1;
		
		VERBOSE_3<<"ERROR: Could not plan for this request: "<<answer[requestType]["returnReport"].asString()<<std::endl;
	}

	VERBOSE_3<<"\t"<<(instance.isGTPAlternative()?"RE":"")<<"Projection "<<(status?"success":"failure")<<", on <"<<newTaskId<<", "<<newTaskAlternativeId<<">, cost="<<newCost<<std::endl;

	instance.setGTPInformation(std::make_pair(newTaskId, newTaskAlternativeId), newCost, planningTime, numberOfAttempt);

	return(status);
#else //MANUAL_DEBUG_MODE
	bool result;
	double cost;
	double planningTime=0.4;
	double numberOfAttempt;
	//Previous action is numbered <-1, -1> to make it obvious that the manual_debug mode is being used
	std::pair<int, int> id(-1,-1);

	if(!instance.isGTPAlternative()){
		std::cout<<"\tResult for GTP projection (0/1) ?   ";
		std::cin>>result;
	}else{
		std::cout<<"\tResult for GTP REprojection (0/1) ?   ";
		std::cin>>result;
	}

//	if(result){
//		std::cout<<"\tAction's cost ?   ";
//                std::cin>>cost;
//	}else{
//		return false;
//	}

	if(result){
		std::cout<<"Id of this new action ?"<<std::endl;
		std::cout<<"ID: "; std::cin>>id.first;
		std::cout<<"Alternative ID: "; std::cin>>id.second;
		
		std::cout<<"\nNumber of attempt (precentage, express in [0, 1]) ?"<<std::endl;
		std::cin>>numberOfAttempt;
	}

//	result=true;
	cost=5;

	instance.setGTPInformation(id, cost, planningTime, numberOfAttempt);

	return result;
#endif //MANUAL_DEBUG_MODE
}

/** Test a geometric function on GTP
 * This is different than a projection since it does not change the geometry
 */
bool ILGTP::testOnGTP(const std::string& taskName, const TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
	
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();

	request.clear();
	request["PlanGTPTask"]["type"]=taskName;

	TaskInstance* predecessor=session.getPredecessorInstance(instance);
	if(predecessor==NULL){
		request["PlanGTPTask"]["previousTaskId"]=-1;
		request["PlanGTPTask"]["previousTaskAlternativeId"]=-1;
	}else{
		request["PlanGTPTask"]["previousTaskId"]=predecessor->getGTPProjection().first;
		request["PlanGTPTask"]["previousTaskAlternativeId"]=predecessor->getGTPProjection().second;

	}
	request["PlanGTPTask"]["computeMotionPlan"]=true;

	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	
	Json::Value answer;
	Json::Reader reader;
	if(!reader.parse(result.second, answer)){
		VERBOSE_3<<"ERROR: Invalid format for answer"<<std::endl;
		return false;
	}
	
	if(answer["PlanGTPTask"].isNull() ||
		!answer["PlanGTPTask"]["returnReport"].isString() ||
		!answer["PlanGTPTask"]["cost"].isDouble() ||
		!answer["PlanGTPTask"]["taskId"].isInt() ||
		!answer["PlanGTPTask"]["taskAlternativeId"].isInt()){
		VERBOSE_3<<"ERROR: Invalid answer to planning request"<<std::endl;
		return false;
	}

	bool status=(answer["PlanGTPTask"]["returnReport"].asString()=="OK");
	VERBOSE_3<<"\tTest "<<(status?"success":"failure")<<", on <"<<request["PlanGTPTask"]["previousTaskId"].asInt()<<", "<<request["PlanGTPTask"]["previousTaskAlternativeId"].asInt()<<">"<<std::endl;

	return (status);
#else //MANUAL_DEBUG_MODE
	bool result;

	std::cout<<"\tResult for GTP test (0/1) ?   ";
	std::cin>>result;

	return result;
#endif //MANUAL_DEBUG_MODE
}

/** Add a constraint on a fact
 * The fact must be true at the end of the GTP Planning for the trajectory to be considered valid
 */
bool ILGTP::addFactConstraint(const std::string& _mainEntity, const std::string& _factType, const std::string& _target, bool value, const TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();

	std::string mainEntity(entityNameMappingToGTP(_mainEntity));
	std::pair<std::string, std::string> type=attributeNameMappingToGTP(_factType);
	//TODO std::string factSubtype(attributeNameMappingToGTP(_factType).second);
	std::string target(entityNameMappingToGTP(_target));

	VERBOSE_3<<"Constraint: Add"<<"("<<mainEntity<<"."<<type.first<<"."<<type.second<<"="<<target<<" - "<<std::boolalpha<<value<<")"<<std::endl;

	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();

	//We create the constraint
	request.clear();
	request["AddGTPConstraintFact"]["mainEntity"]=mainEntity;
	request["AddGTPConstraintFact"]["factType"]=type.first;
	request["AddGTPConstraintFact"]["factSubType"]=type.second;
	request["AddGTPConstraintFact"]["targetEntity"]=target;
	request["AddGTPConstraintFact"]["boolValue"]=value;
//	request["AddGTPConstraintFact"]["valueType"]=true;
//	request["AddGTPConstraintFact"]["specificValue"]=0.0;
	
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);

	return true;
#else //MANUAL_DEBUG_MODE
	std::cerr<<"WARNING: You are using addFactConstraint in debug mode, can not be used."<<std::endl;

	return true;
#endif //MANUAL_DEBUG_MODE
}












////////////////////////////////////////////////////////////////////////////////
//////////////// NON _ GENERIC - PROBLEM SPECIFIC //////////////////////////////
////////////////////////////////////////////////////////////////////////////////


//**************************************//
//	PR2 table-top manipulation      //
//**************************************//
bool ILGTP::pick(const std::string& _agent, const std::string& _object, TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string agent(entityNameMappingToGTP(_agent));
	std::string object(entityNameMappingToGTP(_object));
	
	VERBOSE_3<<"Project: Pick"<<"("<<agent<<", "<<object<<")"<<std::endl;

#ifndef MANUAL_DEBUG_MODE
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();	

	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the agent that picks
	request.clear();
	request["AddGTPAgent"]["key"]="mainAgent";
	request["AddGTPAgent"]["value"]=agent;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the object to pick
	request.clear();
	request["AddGTPObject"]["key"]="mainObject";
	request["AddGTPObject"]["value"]=object;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//TODO REMOVE ONCE BOTH ARMS WORK PROPERLLY
	request.clear();
	request["AddGTPData"]["key"]="hand";
	request["AddGTPData"]["value"]="right";
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
#endif //MANUAL_DEBUG_MODE	
	return projectActionGTP("pick", instance, session);
}

bool ILGTP::place(const std::string& _agent, const std::string& _object, const std::string& _destination, TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string agent(entityNameMappingToGTP(_agent));
	std::string object(entityNameMappingToGTP(_object));
	std::string destination(entityNameMappingToGTP(_destination));
	
	VERBOSE_3<<"Project: Place"<<"("<<agent<<", "<<object<<", "<<destination<<")"<<std::endl;

#ifndef MANUAL_DEBUG_MODE
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();
	
	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the agent that places
	request.clear();
	request["AddGTPAgent"]["key"]="mainAgent";
	request["AddGTPAgent"]["value"]=agent;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the object to place
	request.clear();
	request["AddGTPObject"]["key"]="mainObject";
	request["AddGTPObject"]["value"]=object;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the place where the object ought to go
	request.clear();
	request["AddGTPObject"]["key"]="supportObject";
	request["AddGTPObject"]["value"]=destination;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//TODO REMOVE ONCE BOTH ARM WORK PROPERLLY
	request.clear();
	request["AddGTPData"]["key"]="hand";
	request["AddGTPData"]["value"]="right";
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
#endif //MANUAL_DEBUG_MODE
	
	return projectActionGTP("place", instance, session);
}

bool ILGTP::placeReachable(const std::string& _agent, const std::string& _object, const std::string& _destination, const std::string& _target, TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string agent(entityNameMappingToGTP(_agent));
	std::string target(entityNameMappingToGTP(_target));
	std::string object(entityNameMappingToGTP(_object));
	std::string destination(entityNameMappingToGTP(_destination));
	
	VERBOSE_3<<"Project: PlaceReachable"<<"("<<agent<<", "<<object<<", "<<destination<<", "<<target<<")"<<std::endl;

#ifndef MANUAL_DEBUG_MODE
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();
	
	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the agent that places
	request.clear();
	request["AddGTPAgent"]["key"]="mainAgent";
	request["AddGTPAgent"]["value"]=agent;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
		
	//Set the target agent
	request.clear();
	request["AddGTPAgent"]["key"]="targetAgent";
	request["AddGTPAgent"]["value"]=target;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the object to place
	request.clear();
	request["AddGTPObject"]["key"]="mainObject";
	request["AddGTPObject"]["value"]=object;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the place where the object ought to go
	request.clear();
	request["AddGTPObject"]["key"]="supportObject";
	request["AddGTPObject"]["value"]=destination;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//TODO REMOVE ONCE BOTH ARE WORK PROPERLLY
	request.clear();
	request["AddGTPData"]["key"]="hand";
	request["AddGTPData"]["value"]="right";
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
#endif //MANUAL_DEBUG_MODE

	return projectActionGTP("placeReachable", instance, session);
}

bool ILGTP::stack(const std::string& _agent, const std::string& _object, const std::string& _destination, TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string agent(entityNameMappingToGTP(_agent));
	std::string object(entityNameMappingToGTP(_object));
	std::string destination(entityNameMappingToGTP(_destination));
	
	VERBOSE_3<<"Project: Stack"<<"("<<agent<<", "<<object<<", "<<destination<<")"<<std::endl;

#ifndef MANUAL_DEBUG_MODE
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();
	
	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the agent that places
	request.clear();
	request["AddGTPAgent"]["key"]="mainAgent";
	request["AddGTPAgent"]["value"]=agent;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the object to place
	request.clear();
	request["AddGTPObject"]["key"]="mainObject";
	request["AddGTPObject"]["value"]=object;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the place where the object ought to go
	request.clear();
	request["AddGTPObject"]["key"]="supportObject";
	request["AddGTPObject"]["value"]=destination;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//TODO REMOVE ONCE BOTH ARE WORK PROPERLLY
	request.clear();
	request["AddGTPData"]["key"]="hand";
	request["AddGTPData"]["value"]="right";
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
#endif //MANUAL_DEBUG_MODE

	return projectActionGTP("stackObj", instance, session);
}

bool ILGTP::goTo(const std::string& _agent, const std::string& _destination, TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string agent(entityNameMappingToGTP(_agent));
	std::string destination(entityNameMappingToGTP(_destination));
	
	VERBOSE_3<<"Project: GoTo"<<"("<<agent<<", "<<destination<<")"<<std::endl;

#ifndef MANUAL_DEBUG_MODE
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();
	
	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the agent that moves
	request.clear();
	request["AddGTPAgent"]["key"]="mainAgent";
	request["AddGTPAgent"]["value"]=agent;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the destination
	request.clear();
	request["AddGTPObject"]["key"]="target";
	request["AddGTPObject"]["value"]=destination;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
#endif //MANUAL_DEBUG_MODE
	
	return projectActionGTP("goto", instance, session);
}

bool ILGTP::goToAgent(const std::string& _agent, const std::string& _destination, TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string agent(entityNameMappingToGTP(_agent));
	std::string destination(entityNameMappingToGTP(_destination));
	
	VERBOSE_3<<"Project: GoToAgent"<<"("<<agent<<", "<<destination<<")"<<std::endl;

#ifndef MANUAL_DEBUG_MODE
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();
	
	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the agent that moves
	request.clear();
	request["AddGTPAgent"]["key"]="mainAgent";
	request["AddGTPAgent"]["value"]=agent;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the destination
	request.clear();
	request["AddGTPObject"]["key"]="targetAgent";
	request["AddGTPObject"]["value"]=destination;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
#endif //MANUAL_DEBUG_MODE
	
	return projectActionGTP("goto", instance, session);
}

//TODO Make it less domain specific
bool ILGTP::restPosition(const std::string& _agent, TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string agent(entityNameMappingToGTP(_agent));
	
	VERBOSE_3<<"Project: RestPosition"<<"("<<agent<<")"<<std::endl;

#ifndef MANUAL_DEBUG_MODE
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();
	
	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the agent that moves
	request.clear();
	request["AddGTPAgent"]["key"]="mainAgent";
	request["AddGTPAgent"]["value"]=agent;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the position
	request.clear();
	request["AddGTPData"]["key"]="confNamee";
	request["AddGTPData"]["value"]="restPosi";
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
#endif //MANUAL_DEBUG_MODE
	
	return projectActionGTP("moveTo", instance, session);
}

//TODO Make it less domain specific
bool ILGTP::processPosition(const std::string& _agent, TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string agent(entityNameMappingToGTP(_agent));
	
	VERBOSE_3<<"Project: ProcessPosition"<<"("<<agent<<")"<<std::endl;

#ifndef MANUAL_DEBUG_MODE
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();
	
	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the agent that moves
	request.clear();
	request["AddGTPAgent"]["key"]="mainAgent";
	request["AddGTPAgent"]["value"]=agent;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the position
	//TODO Fix to send a proper position, so far no position is sent because of a hack
#endif //MANUAL_DEBUG_MODE
	
	return projectActionGTP("moveTo", instance, session);
}

bool ILGTP::testVirtualPlace(const std::string& _object, const std::string& _destination, const TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string object(entityNameMappingToGTP(_object));
	std::string destination(entityNameMappingToGTP(_destination));
	
	VERBOSE_3<<"Test: VirtualPlace"<<"("<<object<<", "<<destination<<")"<<std::endl;

#ifndef MANUAL_DEBUG_MODE
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();
	
	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the object to place
	request.clear();
	request["AddGTPObject"]["key"]="mainObject";
	request["AddGTPObject"]["value"]=object;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the place where the object ought to go
	request.clear();
	request["AddGTPObject"]["key"]="supportObject";
	request["AddGTPObject"]["value"]=destination;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
#endif //MANUAL_DEBUG_MODE
	
	return testOnGTP("fakePlace", instance, session);
}

//#ifndef MANUAL_DEBUG_MODE
//TODO FIX bool ILGTP::lock(const std::string& _object, const std::string& _destination, const TaskInstance& instance, const hatpPlanningSession& session) const{
////TODO CALL FAKE PLACE HERE TO CREATE A NODE, GET ITS <ID,ALT> AND USE THEM
//	checkClient();
//	std::string object(entityNameMappingToGTP(_object));
//	std::string destination(entityNameMappingToGTP(_destination));
//	
//	VERBOSE_3<<"Constraint: LockSurface"<<"("<<object<<", "<<destination<<")"<<std::endl;
//	
//	//First we project the virtual object in the right position
//	if(!testVirtualPlace(_object, _destination, instance, session)){
//		return false;
//	}

//	Json::Value request;
//	std::pair<std::string, std::string> result;

//	clearMessageStack();

//	//Clean parameters
//	request.clear();
//	request["ClearGTPInputs"];
//	client.sendMessage(move3D, WRITE(request));
//	result = client.getBlockingMessage();
//	checkStatus(result, VERBOSE_LEVEL);

//	//Once projected, we can call the lock constraint
//	request.clear();
//	request["AddGTPConstraintArea"]["nameId"]="lock";
//	request["AddGTPConstraintArea"]["Object"]=object;
//	request["AddGTPConstraintArea"]["area"]=destination;
//	request["AddGTPConstraintArea"]["isSupport"]=false;
//	client.sendMessage(move3D, WRITE(request));
//	result = client.getBlockingMessage();
//	checkStatus(result, VERBOSE_LEVEL);

//	return(true);
//}
//#endif //MANUAL_DEBUG_MODE

//#ifndef MANUAL_DEBUG_MODE
//TODO FIX bool ILGTP::unlock(const std::string& _object, const std::string& _destination, const TaskInstance& instance, const hatpPlanningSession& session) const{
//	checkClient();
//	std::string object(entityNameMappingToGTP(_object));
//	std::string destination(entityNameMappingToGTP(_destination));
//	
//	VERBOSE_3<<"Constraint: UnlockSurface"<<"("<<object<<", "<<destination<<")"<<std::endl;

//	Json::Value request;
//	std::pair<std::string, std::string> result;

//	clearMessageStack();

//	//Clean parameters
//	request.clear();
//	request["ClearGTPInputs"];
//	client.sendMessage(move3D, WRITE(request));
//	result = client.getBlockingMessage();
//	checkStatus(result, VERBOSE_LEVEL);

//	//Remove the lock constraint
//	request.clear();
//	TaskInstance* predecessor=session.getPredecessorInstance(instance);
//	if(predecessor==NULL){
//		request["RemoveGTPConstFromSol"]["taskId"]=-1;
//		request["RemoveGTPConstFromSol"]["alternativeId"]=-1;
//	}else{
//		request["RemoveGTPConstFromSol"]["taskId"]=predecessor->getGTPProjection().first;
//		request["RemoveGTPConstFromSol"]["alternativeId"]=predecessor->getGTPProjection().second;

//	}
//	client.sendMessage(move3D, WRITE(request));
//	result = client.getBlockingMessage();
//	checkStatus(result, VERBOSE_LEVEL);

//	return(true);
//}
//#endif //MANUAL_DEBUG_MODE

bool ILGTP::paintObject(const std::string& _object, const std::string& _colour, const TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string object(entityNameMappingToGTP(_object));
	
	VERBOSE_3<<"Project: Paint"<<"("<<object<<", "<<_colour<<")"<<std::endl;
	
#ifndef MANUAL_DEBUG_MODE
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();
	
	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	int colour;
	if(_colour=="black" || _colour=="Black" || _colour=="BLACK"){
		colour=0;
	}else if(_colour=="green" || _colour=="Green" || _colour=="GREEN"){
		colour=1;
	}else if(_colour=="blue" || _colour=="Blue" || _colour=="BLUE"){
		colour=2;
	}else if(_colour=="red" || _colour=="Red" || _colour=="RED"){
		colour=3;
	}else if(_colour=="yellow" || _colour=="Yellow" || _colour=="YELLOW"){
		colour=4;
	}else if(_colour=="white" || _colour=="White" || _colour=="WHITE"){
		colour=5;
	}else{
		std::cerr<<"ERROR: Invalid colour given as argument "<<_colour<<std::endl;
		return false;
	}
	
	request.clear();
	request["Change_obj_col"]["object_name"]=object;
	request["Change_obj_col"]["color"]=colour;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
#endif //MANUAL_DEBUG_MODE
	
	return(true);
}




//**************************************//
//		ARCAS			//
//**************************************//
bool ILGTP::pickQuad(const std::string& _agent, const std::string& _object, TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string agent(entityNameMappingToGTP(_agent));
	std::string object(entityNameMappingToGTP(_object));
	
	VERBOSE_3<<"Project: PickQuad"<<"("<<agent<<", "<<object<<")"<<std::endl;

#ifndef MANUAL_DEBUG_MODE	
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();
	
	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the agent that picks
	request.clear();
	request["AddGTPAgent"]["key"]="mainAgent";
	request["AddGTPAgent"]["value"]=agent;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the object to pick
	request.clear();
	request["AddGTPObject"]["key"]="mainObject";
	request["AddGTPObject"]["value"]=object;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
#else
	return true;
#endif //MANUAL_DEBUG_MODE	
	return projectActionGTP("pickquad", instance, session);
}

bool ILGTP::placeQuad(const std::string& _agent, const std::string& _object, const std::string& _destination, TaskInstance& instance, const hatpPlanningSession& session) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string agent(entityNameMappingToGTP(_agent));
	std::string object(entityNameMappingToGTP(_object));
	std::string destination(entityNameMappingToGTP(_destination));
	
	VERBOSE_3<<"Project: PlaceQuad"<<"("<<agent<<", "<<object<<", "<<destination<<")"<<std::endl;

#ifndef MANUAL_DEBUG_MODE
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();
	
	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the agent that places
	request.clear();
	request["AddGTPAgent"]["key"]="mainAgent";
	request["AddGTPAgent"]["value"]=agent;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the object to place
	request.clear();
	request["AddGTPObject"]["key"]="mainObject";
	request["AddGTPObject"]["value"]=object;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the place where the object ought to go
	request.clear();
	request["AddGTPObject"]["key"]="supportObject";
	request["AddGTPObject"]["value"]=destination;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
#else
	return true;
	
#endif //MANUAL_DEBUG_MODE
	
	return projectActionGTP("placequad", instance, session);
}

// TODO CONSTNESS bool ILGTP::pickAndPlaceQuad(const std::string& _agent, const std::string& _object, const std::string& _destination, TaskInstance& instance, const hatpPlanningSession& session) const{
bool ILGTP::pickAndPlaceQuad(const std::string& _agent, const std::string& _object, const std::string& _destination, TaskInstance& instance, const hatpPlanningSession& session){
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE
	std::string agent(entityNameMappingToGTP(_agent));
	std::string object(entityNameMappingToGTP(_object));
	std::string destination(entityNameMappingToGTP(_destination));
	last_projected=object;
	
	VERBOSE_3<<"Project: PickAndPlaceQuad"<<"("<<agent<<", "<<object<<", "<<destination<<")"<<std::endl;

#ifndef MANUAL_DEBUG_MODE
	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();
	
	//Clean parameters
	request.clear();
	request["ClearGTPInputs"];
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the agent that places
	request.clear();
	request["AddGTPAgent"]["key"]="mainAgent";
	request["AddGTPAgent"]["value"]=agent;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the object to place
	request.clear();
	request["AddGTPObject"]["key"]="mainObject";
	request["AddGTPObject"]["value"]=object;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
	
	//Set the place where the object ought to go
	request.clear();
	request["AddGTPObject"]["key"]="supportObject";
	request["AddGTPObject"]["value"]=destination;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();
	checkStatus(result, VERBOSE_LEVEL);
#else
	return true;
	
#endif //MANUAL_DEBUG_MODE
	
	return projectActionGTP("pickplacequad", instance, session);
}
