//The Intermediate Layer for the GTP planner (ILGTP)

//Provides a set of tools to be able to manipulate GTP

#ifndef ILGTP_HH
#define ILGTP_HH

//Uncomment to trigger the manual-debug mode: (almost) every interaction with GTP is turned into user prompt
//#define MANUAL_DEBUG_MODE

#include <iostream>

#include <json/json.h>

#include <msgClient.hh>
#include <TaskInstance.hh>
#include <hatpPlanningSession.hh>
#include <VerboseLevel.hh>

const std::string move3D="move3d";

#include <clientFunctions.hh> //Offers function checkStatus to check if the returned value of a request has a status=OK
//Macro to simplify the writing, necessary to keep constness of ILGTP member functions (getSymbolicFacts, ...)
#define WRITE(x) Json::FastWriter().write(x)
//FastWriter used to send the request with character efficient form

//All those functions are generated when compiling a domain (with GTP enabled)
std::string entityNameMappingToHATP(const std::string& name);
std::string attributeNameMappingToHATP(const std::string& name);
std::string entityNameMappingToGTP(const std::string& name);
std::pair<std::string, std::string> attributeNameMappingToGTP(const std::string& name);
std::vector<std::string> gtpFactsToUse();
//template<class AnyWorldBase> void clearAllGeometricAttributes(AnyWorldBase& wb);
void clearAllGeometricAttributes(WorldBase& wb);
void clearAllGeometricAttributes(LightWorldBase& wb);

class ILGTP {
private:
	//Logging level
	unsigned short VERBOSE_LEVEL;
	
	//MsgConnector
	msgClient client;
	
	//If the list of facts to get is empty, then useGetSymbolicFacts=false so get facts is not called (save some communication time)
	bool useGetSymbolicFacts;

	//Prevents any undesired construction
	ILGTP();
	ILGTP(const ILGTP&);
	void operator=(const ILGTP&);
	
	bool projectActionGTP(const std::string& task, TaskInstance& instance, const hatpPlanningSession& session) const;
	bool testOnGTP(const std::string& taskName, const TaskInstance& instance, const hatpPlanningSession& session) const;
	
	template<class AnyWorldBase> bool applySymbolicFacts(AnyWorldBase& wb, bool getInitialState, const std::string& agentName, std::pair<int, int> stateToGet) const;
	template<class AnyWorldBase> bool addFact(AnyWorldBase& wb, bool getInitialState, const std::string& agentName, const std::string& entityName, const std::string& factType, const std::string& factSubtype, const std::string& newValue, double newSpecificValue) const;
public:
	//Singleton
	static ILGTP& getInstance(){
		static ILGTP gtp;
		
		return gtp;
	}
	~ILGTP();

	//Connect to GTP through TCL
	bool connectGTP(const std::string& server, unsigned int port, unsigned short VERBOSE_LEVEL=0);
	bool disconnectGTP();
	
	//Called before any planning phase is started
	bool initGTP(WorldBase& wb);
	
	//Generic tool to use GTP
	bool addFactConstraint(const std::string& mainEntity, const std::string& factType, const std::string& target, bool value, const TaskInstance& instance, const hatpPlanningSession& session) const;
	
	//TODO
	std::string last_projected;
	
	template<class AnyWorldBase> bool getSymbolicFacts(AnyWorldBase& wb, bool getInitialState, std::pair<int, int> stateToGet) const;
//	
//	
//	
//	
//	//PR2 table-top
//	//Functions specific to PR2 table-top manipulations
	bool pick(const std::string& agent, const std::string& object, TaskInstance& instance, const hatpPlanningSession& session) const;
	bool place(const std::string& agent, const std::string& object, const std::string& destination, TaskInstance& instance, const hatpPlanningSession& session) const;
	bool placeReachable(const std::string& agent, const std::string& object, const std::string& destination, const std::string& targetAgent, TaskInstance& instance, const hatpPlanningSession& session) const;
	bool stack(const std::string& agent, const std::string& object, const std::string& destination, TaskInstance& instance, const hatpPlanningSession& session) const;
	bool goTo(const std::string& agent, const std::string& destination, TaskInstance& instance, const hatpPlanningSession& session) const;
	bool goToAgent(const std::string& agent, const std::string& destinationAgent, TaskInstance& instance, const hatpPlanningSession& session) const;
	//TODO Make them less domain specific
	bool restPosition(const std::string& agent, TaskInstance& instance, const hatpPlanningSession& session) const;
	bool processPosition(const std::string& agent, TaskInstance& instance, const hatpPlanningSession& session) const;
	
//TODO FIX	bool lock(const std::string& object, const std::string& destination, const TaskInstance& instance, const hatpPlanningSession& session) const;
//TODO FIX	bool unlock(const std::string& object, const std::string& destination, const TaskInstance& instance, const hatpPlanningSession& session) const;
	bool paintObject(const std::string& object, const std::string& colour, const TaskInstance& instance, const hatpPlanningSession& session) const;

	bool testVirtualPlace(const std::string& object, const std::string& destination, const TaskInstance& instance, const hatpPlanningSession& session) const;
	
	//ARCAS
	bool pickQuad(const std::string& agent, const std::string& object, TaskInstance& instance, const hatpPlanningSession& session) const;
	bool placeQuad(const std::string& agent, const std::string& object, const std::string& destination, TaskInstance& instance, const hatpPlanningSession& session) const;
//TODO CONSTNESS	bool pickAndPlaceQuad(const std::string& agent, const std::string& object, const std::string& destination, TaskInstance& instance, const hatpPlanningSession& session) const;
	bool pickAndPlaceQuad(const std::string& agent, const std::string& object, const std::string& destination, TaskInstance& instance, const hatpPlanningSession& session);
}; //End of namespace

//Part to integrate it properlly with hatponboard-lib
#define applyEffectsGTP(a, b) ILGTP::getInstance().getSymbolicFacts(a, false, b.getGTPProjection())
#define costGTP() taskInstance.getGTPCost()

//PR2 table-top
//Projections
#define pickGTP(a, b) ILGTP::getInstance().pick(a, b, taskInstance, session)
#define placeGTP(a, b, c) ILGTP::getInstance().place(a, b, c, taskInstance, session)
#define placeReachableGTP(a, b, c, d) ILGTP::getInstance().placeReachable(a, b, c, d, taskInstance, session)
#define stackGTP(a, b, c) ILGTP::getInstance().stack(a, b, c, taskInstance, session)
#define goToGTP(a, b) ILGTP::getInstance().goTo(a, b, taskInstance, session)
#define goToAgentGTP(a, b) ILGTP::getInstance().goToAgent(a, b, taskInstance, session)
//TODO make them less domain specific
#define restPositionGTP(a) ILGTP::getInstance().restPosition(a, taskInstance, session)
#define processPositionGTP(a) ILGTP::getInstance().processPosition(a, taskInstance, session)

//GTP-level actions (no motion path)
//TODO FIX #define lockSurface(a, b) ILGTP::getInstance().lock(a, b, taskInstance, session)
//TODO FIX #define unlockSurface(a, b) ILGTP::getInstance().unlock(a, b, taskInstance, session)
#define paint(a, b) ILGTP::getInstance().paintObject(a, b, taskInstance, session)
#define virtualPlace(a, b) ILGTP::getInstance().testVirtualPlace(a, b, taskInstance, session)

//Allow to add constraint on the geometric search
#define addConstraint(a, b, c, d) ILGTP::getInstance().addFactConstraint(a, b, c, d, taskInstance, session)

//ARCAS
#define pickQuadGTP(a, b) ILGTP::getInstance().pickQuad(a, b, taskInstance, session)
#define placeQuadGTP(a, b, c) ILGTP::getInstance().placeQuad(a, b, c, taskInstance, session)
#define pickAndPlaceQuadGTP(a, b, c) ILGTP::getInstance().pickAndPlaceQuad(a, b, c, taskInstance, session)






//Define the template function of the class (can not be embded in a shared library)
/** Get the symbolic state of GTP
 * @param getInitialState Represent the fact that the inital state is fetched, in this case even static attributes are read
 */
 //Needs to be in the .hh because it uses templates
template<class AnyWorldBase>
inline bool ILGTP::getSymbolicFacts(AnyWorldBase& wb, bool getInitialState, std::pair<int, int> stateToGet) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();
#endif //MANUAL_DEBUG_MODE

	bool status=false;
	if(useGetSymbolicFacts){
	//TODO For belief managment
//		if(wb.isActiveBM()){
//			//Browse database then wbagents if bm enabled
//			for(allAgents
//			VERBOSE_3<<"Get symbolic facts from GTP for agent " << agentName << std::endl;
//			applySymbolicFacts(wb, getInitialState, agentName, stateToGet);
//		}else{
			VERBOSE_3<<"Get symbolic facts from GTP"<<std::endl;
			status=applySymbolicFacts(wb, getInitialState, "", stateToGet);
//		}
	}else{
		status=true;
	}
	//For debug purpose, show the state of the database after the update
//	VERBOSE_3<<wb.toString()<<std::endl;

	return status;
}

/** The actual implementation of fact fetching for an agent
 * @param getInitialState Represent the fact that the inital state is fetched, in this case even static attributes are read
 * @param agent The agent name, can be "" in the case of inactive Belief Management
 */
 //Needs to be in the .hh because it uses templates
template<class AnyWorldBase>
inline bool ILGTP::applySymbolicFacts(AnyWorldBase& wb, bool getInitialState, const std::string& agentName, std::pair<int, int> stateToGet) const{
#ifndef MANUAL_DEBUG_MODE
	checkClient();

	Json::Value request;
	std::pair<std::string, std::string> result;

	clearMessageStack();

	request.clear();
	request["GetGTPFacts"]["taskId"]=stateToGet.first;
	request["GetGTPFacts"]["alternativeId"]=stateToGet.second;
	client.sendMessage(move3D, WRITE(request));
	result = client.getBlockingMessage();

	Json::Value answer;
	Json::Reader reader;
	if(!reader.parse(result.second, answer)){
		VERBOSE_3<<"ERROR: Invalid format for fact message"<<std::endl;
		return false;
	}

	Json::Value facts=answer["GetGTPFacts"]["facts"];
	if(facts.isNull() || !facts.isArray() || facts.empty())
		return false;
	Json::Value::ArrayIndex i;
	for(i=0; i<facts.size(); ++i){
		if(facts[i]["mainEntity"].isNull() || !facts[i]["mainEntity"].isString())
			return false;
		if(facts[i]["factType"].isNull() || !facts[i]["factType"].isString())
			return false;
		if(facts[i]["factSubType"].isNull() || !facts[i]["factSubType"].isString())
			return false;
		if(facts[i]["targetEntity"].isNull() || !facts[i]["targetEntity"].isString())
			return false;
//TODO FIX THIS
		if(facts[i]["specificValue"].isNull() || !facts[i]["specificValue"].isDouble())
			return false;

		std::string mainEntity=facts[i]["mainEntity"].asString();
		std::string type=facts[i]["factType"].asString();
		std::string subType=facts[i]["factSubType"].asString();
		std::string value=facts[i]["targetEntity"].asString();
		double newSpecificValue=facts[i]["specificValue"].asDouble();

//		std::cout<<"ADD FACT: "<<mainEntity<<"."<<type<<"."<<subType<<" = "<<value<<std::endl;
//TODO FIX THIS
		if(! addFact(wb, getInitialState, agentName, mainEntity, type, subType, value, newSpecificValue) ) return false;
	}
	
#endif //MANUAL_DEBUG_MODE
	
	return true;
}

#ifndef MANUAL_DEBUG_MODE
/** Apply every individual fact
 * The subtype of the fact is not considered, all the fact of a given type are combined
 * @param getInitialState Represent the fact that the inital state is fetched, in this case even static attributes are read
 * @param agentName can be empty if not using the belief management
 */
 //Needs to be in the .hh because it uses templates
template<class AnyWorldBase>
inline bool ILGTP::addFact(AnyWorldBase& wb, bool getInitialState, const std::string& agentName, const std::string& entityName, const std::string& factType, const std::string& factSubtype, const std::string& newValue, double newSpecificValue) const{
	checkClient();
	
	bool status=false;
	typedef typename AnyWorldBase::EntityType AnyEntity;
	
	std::string hatpEntityName(entityNameMappingToHATP(entityName));
	AnyEntity* entity=wb.getEntity(agentName, hatpEntityName);
	
	if(entity==NULL){
		VERBOSE_3<<"WARNING: Entity "<<hatpEntityName<<" does not exist"<<(wb.isActiveBM()?" for agent "+agentName:"")<<std::endl;
		return true; //Return True since it is only a warning
	}

	std::string value(entityNameMappingToHATP(newValue));

	//TODO Should be triggered if the fact of a number instead of being an hack
//	if(strcmp(fact.factType.name, "Size")==0){
//		std::stringstream os;
//		os<<newSpecificValue;
//		value=os.str();
//	}
	
	std::string attributeName;
	attributeName=attributeNameMappingToHATP(factType+"."+factSubtype);
	
	if(entity->hasAttributeOfName(attributeName)){
		if(entity->isDynamic(attributeName) || getInitialState){
//			std::cout<<"ADD HATP FACT: "<<hatpEntityName<<"."<<attributeName<<" = "<<value<<std::endl;
			if(entity->isAtom(attributeName)){ //The attrbiute is an atom
				status=wb.updateAtomAttribute(agentName, hatpEntityName, attributeName, value);
			}else{ //The attribute is a set
				status=wb.addToSetAttribute(agentName, hatpEntityName, attributeName, value);
			}
//		}else{
//			VERBOSE_3<<"WARNING: The facts update tried to modifiy "<<hatpEntityName<<"."<<attributeName<<" but it is a static attrribute"<<std::endl;
//			return true; //Return true since it is only a warning
		}
	}else{
//		VERBOSE_3<<"WARNING: Entity "<<hatpEntityName<<" does not have an attribute "<<attributeName<<std::endl;
	}
	
	attributeName=attributeNameMappingToHATP(factType);
	
	if(!entity->hasAttributeOfName(attributeName)){
//		VERBOSE_3<<"WARNING: Entity "<<hatpEntityName<<" does not have an attribute "<<attributeName<<std::endl;
		return true; //Return True since it is only a warning
	}
	
	if(entity->isDynamic(attributeName) || getInitialState){
//		std::cout<<"ADD HATP FACT: "<<hatpEntityName<<"."<<attributeName<<" = "<<value<<std::endl;
		if(entity->isAtom(attributeName)){ //The attrbiute is an atom
			status=wb.updateAtomAttribute(agentName, hatpEntityName, attributeName, value);
		}else{ //The attribute is a set
			status=wb.addToSetAttribute(agentName, hatpEntityName, attributeName, value);
		}
	}else{
//		VERBOSE_3<<"WARNING: The facts update tried to modifiy "<<hatpEntityName<<"."<<attributeName<<" but it is a static attrribute"<<std::endl;
		return true; //Return true since it is only a warning
	}
	return status;
}
#endif //MANUAL_DEBUG_MODE

#endif //ILGTP_HH
