#include<iostream>

#include "LightWorldBaseAgent.hh"

using namespace std;

/** Repalce a LightEntity with the given one.
 * @param sName Name of the LightEntity to replace
 * @param e The new entity
 */
void LightWorldBaseAgent::replaceEntity(std::string const& sName, LightEntity* e) {
	std::map<string, LightEntity*>::iterator it = database.find(sName);
	if(it != database.end()) {
		database.erase(it);
		database.insert(pair<string, LightEntity*>(sName, new LightEntity(*e)));
	}
	else cerr << "ERROR for ReplaceEntity: LightEntity not found in WorldBaseAgent." << endl;

}

/** Check if a LightEntity exists in the agent's database.
 * @param sName Name of the entity to check
 * @return True if the entity exists
 */
bool LightWorldBaseAgent::containEntity(std::string const& sName) const {
	return database.find(sName) != database.end();
}

/** Check if an LightEntity (with a specific type) exists in the agent's database.
 * @param sName Name of the entity to check
 * @param sType The type the entity need to be
 * @return True if the entity exists and is of the given type
 */
bool LightWorldBaseAgent::containEntity(std::string const& sName, std::string const& sType) const {
	map<string, LightEntity*>::const_iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType) return true;
	return false;
}

/** Find a LightEntity by its name and type in the agent's database.
 * @param sName Name of the entity to find
 * @param sType Type of the entity
 * @return Pointer to the entity
 */
LightEntity* LightWorldBaseAgent::find(std::string const& sName, std::string const& sType) {
	map<string, LightEntity*>::iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType) return it->second;
	cerr << "ERROR when trying to access to an unknown Entity \"" << sName << "\" of type \"" << sType;
	cerr << "\" in the WorldBaseAgent." << endl << "NULL returned instead" << endl;
	return NULL;
}

/** Find a LightEntity by its name and type in the agent's database.
 * @param sName Name of the entity to find
 * @param sType Type of the entity
 * @return Const pointer to the entity
 */
LightEntity const* LightWorldBaseAgent::find(std::string const& sName, std::string const& sType) const {
	map<string, LightEntity*>::const_iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType) return it->second;
	cerr << "ERROR when trying to access to an unknown Entity \"" << sName << "\" of type \"" << sType;
	cerr << "\" in the WorldBaseAgent." << endl << "NULL returned instead" << endl;
	return NULL;
}

/** Find a set of all entities matching the given type.
 * @param sType The type the entities need to match
 * @return Set of entities matching the given type
 */
vector<LightEntity*> LightWorldBaseAgent::findByType(string const& sType) const {
	vector<LightEntity*> vec;
	for(map<string,LightEntity*>::const_iterator it=database.begin() ; it!=database.end() ; ++it)
		if(it->second->getType() == sType) vec.push_back(it->second);
	return vec;
}

/** Create a string representing the LightWorldBaseAgent.
 * @return A string representing the agent
 */
string LightWorldBaseAgent::toString() const {
	string s = "***** Start LightWorldBaseAgent Printing *****\n";
	for(map<string,LightEntity*>::const_iterator it=database.begin() ; it!=database.end() ; ++it)
		s += it->second->toString() + "\n";
	s += "***** End LightWorldBaseAgent Printing *****\n";
	return s;
}

/** Return a reference to a LightEntity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param c LightEntity to return
 * @return Reference to the entity, or to an empty entity
 */
LightEntity& LightWorldBaseAgent::operator[](const char* c) {
	std::map<std::string, LightEntity*>::iterator it = database.find(c);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the LightWorldBaseAgent: " << c;
		cerr << " has not been found" << endl;
		cerr << "default Entity returned instead" << endl;
		return defaultLightEntity;
	}
}

/** Return a reference to an LightEntity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param s LightEntity to return
 * @return Reference to the entity, or to an empty entity
 */
LightEntity& LightWorldBaseAgent::operator[](std::string const& s) {
	std::map<std::string, LightEntity*>::iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the LightWorldBaseAgent: " << s;
		cerr << " has not been found" << endl;
		cerr << "default Entity returned instead" << endl;
		return defaultLightEntity;
	}
}

/** Return a reference to an LightEntity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param c LightEntity to return
 * @return Const reference to the entity, or to an empty entity
 */
LightEntity const& LightWorldBaseAgent::operator[](const char* c) const {
	std::map<std::string, LightEntity*>::const_iterator it = database.find(c);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the LightWorldBaseAgent: " << c;
		cerr << " has not been found" << endl;
		cerr << "default Entity returned instead" << endl;
		return defaultLightEntity;
	}
}

/** Return a reference to an LightEntity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param s LightEntity to return
 * @return Const reference to the entity, or to an empty entity
 */
LightEntity const& LightWorldBaseAgent::operator[](std::string const& s) const {
	std::map<std::string, LightEntity*>::const_iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the LightWorldBaseAgent: " << s;
		cerr << " has not been found" << endl;
		cerr << "default Entity returned instead" << endl;
		return defaultLightEntity;
	}
}

/** Destructor.
 * Delete all Entites in the database.
 */
LightWorldBaseAgent::~LightWorldBaseAgent() {
	for(map<string, LightEntity*>::iterator it=database.begin() ; it!=database.end() ; ++it)
		delete it->second;
}
